//------------------------------------------------------------------------------
//
//  npplot.c
//  npe
//
//  Created by Shane Saxon on 2/19/10.
//  Copyright Saxon Digital 2010. All rights reserved.
//
//------------------------------------------------------------------------------


#include "npplot.h"

#include <stdlib.h>

#include "../npctrl.h"



void InitPlotSurface (NPplotPtr plot);
void InitPlotPoints (NPplotPtr plot);
void InitPlotPin (NPplotPtr plot);
void InitPlotVideo (NPplotPtr plot);
void InitPlotGrid (NPplotPtr plot);


//------------------------------------------------------------------------------
void npInitPlot (void* dataRef)
{
	int i = 0;

	pData data = (pData) dataRef;

	for (i=0; i < kNPplotMax; i++)
		data->map.plot[i] = NULL;
}


//------------------------------------------------------------------------------
void npClosePlot(void* dataRef)
{
	int i = 0;

	pData data = (pData) dataRef;

	for (i=0; i < data->map.plotCount; i++)
		npPlotDelete (data->map.plot[i], dataRef);
}


// should make this static and add locking for delete to be thread safe,      debug zz
//------------------------------------------------------------------------------
void* npPlotNew (int plotType, void* dataRef)
{
	int plotIndex = 0;
	NPplotPtr plot = NULL;

	pData data = (pData) dataRef;

	//return if at the plot max
	if ( data->map.plotCount == kNPplotMax )
		return NULL;							// add error plotmax , debug zz

	//init the plot
	plot = (NPplotPtr) malloc (sizeof(NPplot));
	if (data == NULL)
	{
		printf ("\n 4219 error malloc failed cannot write file \n");
		return NULL;
	}

	plot->id = npNewNodeID();
	npInitPlotDefault (plot);	

	plot->type = plotType;				//check to make sure inbounds, debug, zz

	switch (plotType)
	{
		case kPlotCamera :	npInitPlotCamera (plot); break;
		case kPlotGrid :	InitPlotGrid (plot); break;
		case kPlotVideo :	InitPlotVideo (plot); break;
		case kPlotSurface :	InitPlotSurface (plot); break;
		case kPlotPoints :	InitPlotPoints (plot); break;
		case kPlotPin :		
			InitPlotPin (plot); 
			npPlotNewChild (plot, dataRef);	// should not need this, w/o breaks bad, debug, zz
			plot->translate.x += 5.0f * data->map.plotCount -35.0f;
			break;
		case kPlotPinTorus :
			InitPlotPin (plot);
			plot->surfaceType = kSurfaceTorus;
			return (void*)plot;
			break;

		default : 
			break;
	}

	data->map.plot[data->map.plotCount] = plot;
	data->map.plotCount++;
	data->map.plotIndex = data->map.plotCount - 1;


	data->map.activePlot = plot;
	data->map.selectedPlot = data->map.activePlot;
	data->map.selectedIndex = data->map.plotIndex;

	return (void*)plot;
}

					// add a primary torus child to the trunk,
					//					make this part of the npPlotNew.. debug, zz
//------------------------------------------------------------------------------
void* npPlotNewChild (NPplotPtr plot, void* dataRef)
{
	NPplotPtr plotChild = NULL;

	// we only allow creating children of primary toroid,				debug, zz
	if (plot->branchLevel < 1 && plot->childCount > 0)		// allow for multiple primaries
		plot = plot->child[plot->childIndex];

	// create and add child to level 1
	if (plot->childCount >= kNPplotChildMax)
		return NULL;

	plotChild = npPlotNew (kPlotPinTorus, dataRef);
					
	plotChild->parent = plot;
	plotChild->branchLevel = plot->branchLevel + 1;

	// distribute the children at 25 deg spacing... is 0.44 radians
	plotChild->rotateRad.x = 0.44f * (GLfloat)plot->childCount;

	plotChild->shaderType = kNPglutSolidTorus;

	// attach child to parent
	plot->child[plot->childCount] = plotChild;
	plot->childCount++;
	plot->childIndex = plot->childCount - 1;

	return (void*)plotChild;
}

															// add travers tree delete...
//------------------------------------------------------------------------------
void npPlotDelete (NPplotPtr plot, void* dataRef)
{
	int i = 0;

	pData data = (pData) dataRef;
	NPplotPtr parent = NULL;

	if (data->map.plotIndex < 3 || plot == NULL)	//do not delete the root grid or camera
		return;

	if (plot->branchLevel > 0)
	{
		parent = plot->parent;

		// don't delete the last torus...			// fix thi
		if (plot->branchLevel == 1)
		{
			data->map.selectedPlot = parent;
			data->map.activePlot = data->map.selectedPlot;
			return;
		}

		// compact the gap, optimize with block copy later...		zz
		parent->childCount--;
		for (i = parent->childIndex; i < parent->childCount; i++)
			parent->child[i] = parent->child[i+1];

		parent->child[parent->childCount] = NULL;

		if (parent->childIndex >= parent->childCount
			&& parent->childCount > 0 )
			parent->childIndex = parent->childCount - 1;

		if (parent->childCount <= 0)
		{
			data->map.selectedPlot = parent;
			data->map.activePlot = data->map.selectedPlot;
		}
		else
		{
			data->map.selectedPlot = parent->child[parent->childIndex];
			data->map.activePlot = data->map.selectedPlot;
		}
	}
	else
	{
		// compact the gap, optimize with block copy later...		zz
		data->map.plotCount--;
		for (i = data->map.plotIndex; i < data->map.plotCount; i++)
			data->map.plot[i] = data->map.plot[i+1];

		data->map.plot[data->map.plotCount] = NULL;

		if (data->map.plotIndex >= data->map.plotCount)
			data->map.plotIndex = data->map.plotCount - 1;

		switch (plot->type)
		{
			case kPlotCamera :	free ((NPcameraPtr) plot->data); break;
			case kPlotVideo :	free ((NPvideoPtr) plot->data); break;
			case kPlotSurface : free ((NPsurfacePtr) plot->data); break;
			case kPlotPoints :	free ((NPpointsPtr) plot->data); break;
			case kPlotPin :		free ((NPpinPtr) plot->data); break;
			case kPlotGrid :	free ((NPgridPtr) plot->data); break;
			default : 
				break;
		}
		data->map.selectedPlot = data->map.plot[data->map.plotIndex];
		data->map.activePlot = data->map.selectedPlot;
	}

	// if last plot deleted then select the camera, note perhaps re-order grid and camera, debug zz
	if ( data->map.plotIndex < 3)
	{
		data->map.plotIndex = 1;
		data->map.selectedPlot = data->map.plot[data->map.plotIndex];
		data->map.activePlot = data->map.selectedPlot;
	}

	free (plot);
}




//------------------------------------------------------------------------------
void npInitPlotDefault (NPplotPtr plot)
{
	int i = 0;

//	plot->id = npNewNodeID();	// called by npNew and preserved during reset

	plot->type			= kPlotDefault;		//plot type, point, pin, cam, video

	plot->data			= NULL;				//plot type specific data

	plot->selected		= 0;				//defualt is not selected

	plot->parent		= NULL;				//parent plot, binary tree of plots
	plot->branchLevel	= 0;			//0 is the trunk pin (ice-cream cone)

	for (i=0; i < kNPplotChildMax; i++)
		plot->child[i] = NULL;

	plot->childIndex	= 0;
	plot->childCount	= 0;				//no children by default

	plot->channel.x		= kChannelNull;		//channel to use for x,y,z points
	plot->channel.y		= kChannelNull;
	plot->channel.z		= kChannelNull;
	
	plot->channelIndex.x = 0;				//last updated channel index
	plot->channelIndex.y = 0;
	plot->channelIndex.z = 0;

	plot->averageType.x	= kAverageNull;		//averaging type applied to data
	plot->averageType.y	= kAverageNull;
	plot->averageType.z	= kAverageNull;
	
	plot->sampleInterval.x = 1;				//the interval to be averaged
	plot->sampleInterval.y = 1;
	plot->sampleInterval.z = 1;

	plot->rotate.angle	= 0.0f;				//orientation vector
	plot->rotate.x		= 0.0f;
	plot->rotate.y		= 0.0f;
	plot->rotate.z		= 1.0f;
	
	plot->scale.x		= 1.0f;				//scale plot XYZ
	plot->scale.y		= 1.0f;
	plot->scale.z		= 1.0f;
	
	plot->translate.x	= 0.0f;				//location relative to plotOrigin
	plot->translate.y	= 0.0f;
	plot->translate.z	= 0.0f;

	plot->plotOrigin.x	= 0.0f;				//origin relative to parent origin
	plot->plotOrigin.y	= 0.0f;
	plot->plotOrigin.z	= 0.0f;
	
	plot->rotateRate.x	= 0.0f;
	plot->rotateRate.y	= 0.0f;
	plot->rotateRate.z	= 0.0f;

	plot->rotateRad.x	= 0.0f;
	plot->rotateRad.y	= 0.0f;
	plot->rotateRad.z	= 0.0f;
	
	plot->scaleRate.x	= 0.0f;				//scale rate 
	plot->scaleRate.y	= 0.0f;
	plot->scaleRate.z	= 0.0f;

	plot->translateRate.x = 0.0f;			//pans the location of the plot, pronunciar en Espanol, lol
	plot->translateRate.y = 0.0f;
	plot->translateRate.z = 0.0f;

	plot->translateVect.x = 0.0f;			//pans the location of the plot, pronunciar en Espanol, lol
	plot->translateVect.y = 0.0f;
	plot->translateVect.z = 0.0f;
	
	plot->surfaceType	= 1;//kSurfaceNull;		//FFT layers, mesh, sphere...
//	plot->primitive		= 1;														//zz debug
	plot->shaderType	= 1;		//wire, flat, gouraud, phong

	plot->lineWidth		= 2.0f;				//draws lines, zero for off
	plot->pointSize		= 0.0f;				//draws dots, zero for off
	plot->circle		= 0.0f;				//draw circle at current data point

	plot->colorIndex	= 0;
	
	plot->color.r		= 50;				//default color assigned to new data
	plot->color.g		= 101;
	plot->color.b		= 101;
	plot->color.a		= 255;				//fully opaque

	plot->colorFade		= false;

	plot->hide			= false;			//hide the plot, data stays active
	plot->freeze		= false;			//freezes data and physics updates
	plot->center		= false;			//centers to plotOrigin
	plot->centerPerGrid	= false;			//centers in gridSpace increments

	plot->autoZoom.x	= false;			//scales plot to fit screen
	plot->autoZoom.y	= false;
	plot->autoZoom.z	= false;

	plot->scroll		= false;			//scroll right to left
	
	plot->triggerOnHi.x	= false;			//triggers are multipurpose
	plot->triggerOnHi.y	= false;
	plot->triggerOnHi.z	= false;

	plot->triggerOnLow.x = false;
	plot->triggerOnLow.y = false;
	plot->triggerOnLow.z = false;
	
	plot->triggerLevelHi.x = 0.0f;
	plot->triggerLevelHi.y = 0.0f;
	plot->triggerLevelHi.z = 0.0f;

	plot->triggerLevelLow.x = 0.0f;
	plot->triggerLevelLow.y = 0.0f;
	plot->triggerLevelLow.z = 0.0f;

	plot->triggerCommand = kCtrlCommandNull;	//execute ctrlCommand upon trigger
	
	plot->proximity.x	= false;			//used for collision engine 
	plot->proximity.y	= false;
	plot->proximity.z	= false;
	
	plot->proximityRadius.x = 0.0f;
	plot->proximityRadius.y = 0.0f;
	plot->proximityRadius.z = 0.0f;

	plot->drawCurrentValue.x = false;		//draw current values as text
	plot->drawCurrentValue.y = false;
	plot->drawCurrentValue.z = false;

	plot->selectionSet.x	= false;		//default to no selection set
	plot->selectionSet.y	= false;
	plot->selectionSet.z	= false;

	plot->size = sizeof(NPplot);
}


// cameras can be GL scene cameras or real-world cameras
// video is attached as a daughter of the recPlot
//-----------------------------------------------------------------------------
void npInitPlotCamera (NPplotPtr plot)
{
	NPcameraPtr data = (NPcameraPtr) malloc (sizeof(NPcamera));
	if (data == NULL)
	{
		printf ("\n 4215 error malloc failed cannot write file \n");
		return;
	}

	plot->data  = (void*) data;

	plot->type			= kPlotCamera;

	data->id			= plot->id;
	data->type			= kPlotCamera;

	data->textureID		= 0;					//texture ID in GL


	data->format		= kVideoFormat720p;
	data->interlaced	= kVideoFieldNull;		//progressive scan, no fields
	data->stereo		= kVideoStereoNull;		//not stereoscopic
	data->aspectRatio	= 1.77778f;				//1.77778 is approx 16:9
	data->colorSpace	= kColorSpaceRGB_8;		//8bit RGB is what GPU's output
	data->fps			= 59.97f;
	data->width			= 1280;
	data->height		= 720;

	data->fov			= 35.0f;				//35mm, 70mm...
	data->clipNear		= 1.0f;					//clipping planes
	data->clipFar		= 2000.0f;				//increased clipping plane, ejs

	data->aperture		= 1.0f;					//F stop
	data->exposure		= 0.01f;				//in seconds
	data->sensorType	= kCameraSensorNull;	//3CCD, Debayer pattern...

	data->size = sizeof(NPcamera);
}


//-----------------------------------------------------------------------------
void InitPlotVideo (NPplotPtr plot)
{
	NPvideoPtr data = (NPvideoPtr) malloc (sizeof(NPvideo));
	if (data == NULL)
	{
		printf ("\n 4214 error malloc failed cannot write file \n");
		return;
	}

	plot->data  = (void*) data;

	plot->type			= kPlotVideo;

	data->id			= plot->id;
	data->type			= kPlotVideo;

	data->textureID		= 0;

	data->format		= kVideoFormat720p;
	data->interlaced	= kVideoFieldNull;		//progressive scan, no fields
	data->stereo		= kVideoStereoNull;		//not stereoscopic
	data->aspectRatio	= 1.77778f;				//1.77778 is approx 16:9
	data->colorSpace	= kColorSpaceRGB_8;		//8bit RGB is what GPU's output
	data->fps			= 59.97f;
	data->width			= 1280;
	data->height		= 720;

	data->bias.r		= 0.0f;				//brightness shift, 0.0 no change
	data->bias.g		= 0.0f;
	data->bias.b		= 0.0f;
	data->bias.a		= 0.0f;
	data->gain.r		= 1.0f;				//color multiplier, 1.0 no change
	data->gain.g		= 1.0f;
	data->gain.b		= 1.0f;
	data->gain.a		= 1.0f;

	data->size = sizeof(NPvideo);
}


//------------------------------------------------------------------------------
void InitPlotSurface (NPplotPtr plot)
{
	NPsurfacePtr data = (NPsurfacePtr) malloc (sizeof(NPsurface));
	if (data == NULL)
	{
		printf ("\n 4215 error malloc failed cannot write file \n");
		return;
	}

	plot->data  = (void*) data;

	plot->type			= kPlotSurface;

	data->id			= plot->id;
	data->type			= kPlotSurface;

	data->textureID		= 0;

	data->vertexArray	= NULL;
	data->normalArray	= NULL;
	data->colorArray	= NULL;
	data->vertexCount	= 0;
	data->normalCount	= 0;
	data->arrayIndex	= 0;

	data->size = sizeof(NPsurface);
}


//-----------------------------------------------------------------------------
void InitPlotPoints (NPplotPtr plot)
{
	NPpointsPtr data = (NPpointsPtr) malloc (sizeof(NPpoints));
	if (data == NULL)
	{
		printf ("\n 4216 error malloc failed cannot write file \n");
		return;
	}

	plot->data  = (void*) data;
	
	plot->type			= kPlotPoints;

	data->id			= plot->id;
	data->type			= kPlotPoints;

	data->vertexArray	= NULL;				//the array of data points
	data->colorArray	= NULL;				//each point gets color assigned
	data->vertexCount	= 0;				//number of points in the plot
	data->arrayIndex	= 0;				//current vertex index in the array

	data->size = sizeof(NPpoints);
}


//-----------------------------------------------------------------------------
void InitPlotPin (NPplotPtr plot)
{
	NPpinPtr data = (NPpinPtr) malloc (sizeof(NPpin));
	if (data == NULL)
	{
		printf ("\n 4217 error malloc failed cannot write file \n");
		return;
	}

	plot->data  = (void*) data;


	plot->type			= kPlotPin;
	plot->surfaceType	= kSurfacePin;		//ice-cream cone shape
	plot->shaderType		= kNPprimitivePin;							//zz debug, zz

	data->id			= plot->id;
	data->type			= kPlotPin;

	data->textureID		= 0;				//0 for no texture assigned

	data->innerRadius	= 0.15f;			//inner radius of the toroid
	data->outerRadius	= 1.5f;			//outer radius of the toroid

	// 4 * outerRadius / innerRadius -> int
	data->radiusRatioIndex = 40;			//sorts pre-loaded GPU toriods
	data->screenSizeIndex = 8;				//same as above for detail level

	data->slices		= 7;				//calculated from screenSizeIndex
	data->stacks		= 16;

	data->translateTex.x = 0.0f;			//position offset of texture map
	data->translateTex.y = 0.0f;
	data->translateTex.z = 0.0f;

	data->scaleTex.x	= 1.0f;
	data->scaleTex.y	= 1.0f;
	data->scaleTex.z	= 1.0f;

	data->rotateTex.x	= 0.0f;
	data->rotateTex.y	= 0.0f;
	data->rotateTex.z	= 0.0f;
	data->rotateTex.angle = 0.0f;			//orientation of texture map

	data->size = sizeof(NPpin);
}


//------------------------------------------------------------------------------
void InitPlotGrid (NPplotPtr plot)
{
	NPgridPtr data = (NPgridPtr) malloc (sizeof(NPgrid));
	if (data == NULL)
	{
		printf ("\n 4218 error malloc failed cannot write file \n");
		return;
	}

	plot->data  = (void*) data;

	plot->type			= kPlotGrid;

	plot->colorIndex	= 3;
	plot->color.r		= 0;
	plot->color.g		= 0;
	plot->color.b		= 255;
	plot->color.a		= 255;

	data->id			= plot->id;
	data->type			= kPlotGrid;

	data->segments.x	= 16;
	data->segments.y	= 16;
	data->segments.z	= 0;
	data->length.x		= 22.6f;
	data->length.y		= 11.2f;
	data->length.z		= 10.0f;

	data->overlay		= false;

	data->size			= sizeof(NPgrid);
}

